@@ -49,7 +49,7 @@ class JobStatus(Enum):
 def get_env(val, env, default=None):
     if val is None:
         env_val = os.getenv(env)
-
+        #log.debug(env + ': '+str(val))
         if env_val is None:
             if default is None:
                 raise KeyError(f'ENV var <{env}> is not set.')
@@ -60,9 +60,13 @@ def get_env(val, env, default=None):
     else:
         return val
 
+def datetime_to_json_converter(o):
+    if isinstance(o, datetime):
+        return o.__str__()
+
 tzl = pytz.utc
 
-log = logging.getLogger('numerous_client')
+log = logging.getLogger('numerous_client.main')
 log.setLevel(logging.DEBUG)
 
 
@@ -76,7 +80,7 @@ signal.signal(signal.SIGTERM, handle_term)
 
 
 class NumerousBufferedWriter:
-    def __init__(self, write_method, writer_closed_method, scenario:str=None, buffer_size:int=24):
+    def __init__(self, write_method, writer_closed_method, scenario:str=None, buffer_size:int=24*7):
         self.scenario = scenario
         self._tags = None
         self.ix = 0
@@ -230,7 +234,7 @@ class NumerousLogHandler:
 
 
 class NumerousClient:
-    def __init__(self, job_id=None, project=None, scenario=None, server=None, port=None, refresh_token=None, clear_data=None, no_log=False, instance_id=None, secure=None, url=None):
+    def __init__(self, job_id=None, project=None, scenario=None, server=None, port=None, refresh_token=None, clear_data=None, no_log=False, instance_id=None, secure=None, url=None, resumed=None):
         if url is not None:
             from urllib.parse import urlparse
             parsed_url = urlparse(url)
@@ -283,8 +287,8 @@ class NumerousClient:
 
         clear_data = get_env(clear_data, 'NUMEROUS_CLEAR_DATA', default='False') == 'True'
 
-        # If clearing data, the job was not resumed and vice versa
-        self._was_resumed = not clear_data
+        self._was_resumed = get_env(resumed, 'NUMEROUS_JOB_RESUMED', default='False') == 'True'
+
 
         self._hibernate_callback = self._default_hibernation_callback
         self._hibernate_callback_args = ([], {})
@@ -317,13 +321,13 @@ class NumerousClient:
 
         log.info(f"Client connecting to: {server}:{port}, using secure channel: {secure_channel}")
         if str(secure_channel) == 'True':
-            cert = str.encode(get_cert(f'https://{server}:4443/cert'))
-            creds = grpc.ssl_channel_credentials(cert)
+            #cert = str.encode(get_cert(f'https://{server}:4443/cert'))
+            creds = grpc.ssl_channel_credentials()
             # Deal with cert being self-signed:
-            cert_decoded = x509.load_pem_x509_certificate(cert, default_backend())
-            cert_cn = cert_decoded.subject.rfc4514_string().split('CN=')[-1].split(',')[0]
-            options = (('grpc.ssl_target_name_override', cert_cn),)
-            channel = grpc.secure_channel(f'{server}:{port}', creds, options)
+            #cert_decoded = x509.load_pem_x509_certificate(cert, default_backend())
+            #cert_cn = cert_decoded.subject.rfc4514_string().split('CN=')[-1].split(',')[0]
+            #options = (('grpc.ssl_target_name_override', cert_cn),)
+            channel = grpc.secure_channel(f'{server}:{port}', creds)
         else:
             channel = grpc.insecure_channel(f'{server}:{port}')
 
@@ -406,6 +410,8 @@ class NumerousClient:
 
 
     def get_scenario(self, scenario=None, project=None, path='.'):
+        os.makedirs(path, exist_ok=True)
+
         log.debug('Get scenario')
         if scenario is None:
             scenario = self._scenario
@@ -537,7 +543,8 @@ class NumerousClient:
                                            exception_object_message=exception_object_message,
                                            full_traceback=full_traceback))
 
-    def get_download_files(self, files, local_path='.', scenario=None):
+    def get_download_files(self, files, local_path='./tmp', scenario=None):
+        os.makedirs(local_path, exist_ok=True)
         if scenario is None:
             scenario = self._scenario
     #    project = "Xcqmc04Ckc9y7IbG6BWa"
@@ -582,23 +589,22 @@ class NumerousClient:
 
         upload()
 
-    def get_timeseries_stats(self, project=None, scenario=None, execution=None):
+    def get_timeseries_stats(self, project=None, scenario=None, execution=None, tag='_index'):
         if project is None:
             project = self._project
         if scenario is None:
             scenario = self._scenario
         if execution is None:
             execution = self._execution_id
-
-        return self._base_client.GetScenarioDataStats(spm_pb2.Scenario(project=project, scenario=scenario, execution=execution))
+        print('stat tag: ', tag)
+        return self._base_client.GetScenarioDataStats(spm_pb2.ScenarioStatsRequest(project=project, scenario=scenario, execution=execution, tag=tag))
 
     def get_dataset_closed(self, project=None, scenario=None, execution=None):
         if project is None:
             project = self._project
         if scenario is None:
             scenario = self._scenario
-        if execution is None:
-            execution = self._execution_id
+
 
         return self._base_client.GetDataSetClosed(spm_pb2.Scenario(project=project, scenario=scenario, execution=execution)).is_closed
 
@@ -610,7 +616,7 @@ class NumerousClient:
 
         return self._base_client.GetScenarioMetaData(spm_pb2.Scenario(project=project, scenario=scenario, execution=execution))
 
-    def set_timeseries_meta_data(self, tags:list, aliases:dict={}, offset=0, timezone='UTC', epoch_type='s', scenario= None, execution=None):
+    def set_timeseries_meta_data(self, tags:list, aliases:dict={}, offset=0, timezone='UTC', epoch_type='s', scenario= None, execution=None, project=None):
 
         def g(tag, key, def_):
 
@@ -636,14 +642,15 @@ class NumerousClient:
             return spm_pb2.Tag(
                 **kwargs
             )
-
+        if project is None:
+            project = self._project
         if scenario is None:
             scenario = self._scenario
         if execution is None:
             execution = self._execution_id
 
         return self._base_client.SetScenarioMetaData(
-            spm_pb2.ScenarioMetaData(scenario=scenario, execution=execution,
+            spm_pb2.ScenarioMetaData(project=project, scenario=scenario, execution=execution,
             tags=[wrap_tag(tag) for tag in tags],
             aliases=[spm_pb2.Alias(tag=k, aliases=v) for k,v in aliases.items()],
                                      offset=offset, timezone=timezone, epoch_type=epoch_type
@@ -786,7 +793,7 @@ class NumerousClient:
             for r in l.data:
                 yield r.tag, r.values
 
-    def read_data_as_row(self, tags=[], start:int=0, end:int=0, scenario=None, execution=None, subscribe=False, offset=0, project=None):
+    def read_data_as_row(self, tags=[], start:int=0, end:int=0, scenario=None, execution=None, subscribe=False, offset=0, project=None, time_range=True):
         if project is None:
             project = self._project
 
@@ -797,20 +804,28 @@ class NumerousClient:
         if execution is None:
             execution = self._execution_id
 
-        time_range = (start > 0 or end > 0)
-        log.debug(f'Time range?: {time_range}')
-        end = end if end > 0 else 3600*24*365*10000
-
+        print(project)
+        print(scenario)
+        print(execution)
+        print(tags)
         print('start: ', start)
         print('end: ', end)
 
+        #time_range = (start > 0 or end > 0)
+        #log.debug(f'Time range?: {time_range}')
+        end = end if end > 0 or not time_range else 3600*24*365*10000
+
+
+        print('timerange: ', time_range)
+        print('listen: ', subscribe)
+
         read_data = self._base_client.ReadData(
             spm_pb2.ReadScenario(project=project, scenario=scenario, execution=execution, tags=tags, start=start, end=end, time_range=time_range, listen=subscribe))
         current_row = {}
 
 
         for b in read_data:
-            #print('b: ', b)
+
 
             for d in b.data:
                 if d.tag != "":
@@ -819,9 +834,19 @@ class NumerousClient:
                     else:
                         current_row[d.tag] += list(d.values)
 
+
             if b.row_complete or b.block_complete:
-                #print(current_row.keys())
+                print(current_row.keys())
                 #print(tags)
+                if '_index' in current_row:
+                    len_current = len(current_row)-1
+                else:
+                    len_current = len(current_row)
+
+                log.debug('Meta: '+str(self._base_client.GetScenarioMetaData(
+                    spm_pb2.Scenario(project=project, scenario=scenario, execution=execution))))
+
+                #assert len(current_row)-1 == len(tags), f'Missing tags from result: {set(tags).difference(set(current_row))}'
                 len_index = len(current_row['_index'])
 
                 wrong_len_tags = []
@@ -912,16 +937,30 @@ class NumerousClient:
     def _rows_to_blocks(self, rows):
         blocks = []
         data = {}
+        counter = 0
+
+        def flush_block(data):
+
+            blocks.append([spm_pb2.DataBlock(tag=k, values=v) for k, v in data.items()])
+
         for r in rows:
+            row_count = 0
             for k, v in r.items():
                 if not k in data:
                     data[k] = []
                 data[k].append(v)
+                row_count += 1
 
-        for k, v in data.items():
-
-            blocks.append(spm_pb2.DataBlock(tag=k, values=v))
+            counter += row_count
+            if counter + row_count > 3e6/8:
+                flush_block(data)
+                data = {}
+                counter = 0
 
+        if counter>0:
+            flush_block(data)
+            data = {}
+            counter = 0
 
         return blocks
 
@@ -948,15 +987,15 @@ class NumerousClient:
 
                     blocks = self._rows_to_blocks(rows)
                     rows = []
-
-                    yield spm_pb2.DataList(scenario=scenario, execution=execution, data=blocks, clear=clear, block_complete=False, row_complete=True)
+                    for block in blocks:
+                        yield spm_pb2.DataList(scenario=scenario, execution=execution, data=block, clear=clear, block_complete=False, row_complete=True)
                     clear = False
 
             #if len(rows)>0:
             blocks = self._rows_to_blocks(rows)
 
-
-            yield spm_pb2.DataList(scenario=scenario, execution=execution, data=blocks, clear=clear, block_complete=True, row_complete=True)
+            for block in blocks:
+                yield spm_pb2.DataList(scenario=scenario, execution=execution, data=block, clear=clear, block_complete=True, row_complete=True)
 
             #return
 
@@ -1114,18 +1153,13 @@ class NumerousClient:
         self._access_token_refresher.stop()
 
         # If job has been hibernated, call the set hibernate callback
-
-
         if self._hibernating:
-            print("hibernate...")
             self._hibernate_callback(*self._hibernate_callback_args[0], **self._hibernate_callback_args[1])
 
         if self._complete:
-
-            if not self._hibernating:
-                if not self._error:
-                    print("setting status...")
-                    self.set_scenario_progress(message='Finished', status='finished', force=True)
+            if not self._hibernating and not self._error:
+                print("setting status...")
+                self.set_scenario_progress(message='Finished', status='finished', force=True)
 
             print("complete...")
             self.complete_execution()
@@ -1152,6 +1186,7 @@ def open_client(job_id=None, project=None, scenario=None, clear_data=None, no_lo
         print(f"CLIENT FAILED WITH ERROR CODE: {e} | {traceback.format_exc()}")
         numerous_client_.push_scenario_error(error=traceback.format_exc())
 
+
     finally:
 
         logging.warning("Closing!")
@@ -1209,7 +1244,7 @@ class DynamicSource:
 
         #If data set is open, we will try to listen for data as its being generated if t end not reached.
         self.subscribe = not self.closed
-        continuous = not self.subscribe
+        continuous = True#not self.subscribe
 
         self.offset = offset
         #self.meta.offset + self.spec['offset']
@@ -1238,8 +1273,8 @@ class DynamicSource:
         print('start: ', start)
         print('end: ', end)
 
-        tags=[s['tag'] for s in spec['tags'].values()]
-
+        tags=set([s['tag'] for s in spec['tags'].values()])
+        log.debug('Datasource Tags: '+str(tags))
         self.generator = client.read_data_as_row(scenario=scenario, tags=tags,
                                                  execution=execution, start=start, end=end,
                                                  subscribe=self.subscribe)
@@ -1312,7 +1347,11 @@ class DynamicSource:
 
         row = {}
         for tag, s in self.spec['tags'].items():
-            row[tag] = row_[s['tag']]*s['scale'] + s['offset']
+            try:
+                row[tag] = row_[s['tag']]*s['scale'] + s['offset']
+            except:
+                log.debug(str(row_))
+                raise
 
         #print('offset: ', self.offset)
 
@@ -1338,7 +1377,8 @@ class DynamicSource:
             self.row_buffer.trim()
 
         while self.t <= t:
-
+            #print(t)
+            #print(self.t)
             part_row = self.__next__()
 
             #print('prd: ', part_row)
@@ -1350,7 +1390,7 @@ class DynamicSource:
 import numpy as np
 class RowBuffer:
     def __init__(self, tags=None, offset=None):
-        self.buffer_len = 5
+        self.buffer_len = 100
         self._buffer = None
         self._pos = 0
         self.tags = tags
@@ -1451,34 +1491,27 @@ class InputManager:
                 else:
                     raise e
 
-            row.update(part_row)
 
-        row.update({'_index': t, '_index_relative': t - self.t0})
+            #if isinstance(part_row, dict):
+            #    tolerance = 0
+            #    if '_index_relative' in row and '_index_relative' in part_row and abs(row['_index_relative'] - part_row['_index_relative'])>tolerance:
 
-        return row
-
-    def __next__(self):
-
-        row = {}
-
-        # print('t: ', t)
-        for source in self.sources:
-
-            try:
-                part_row = source.__next__()
-
-            except TimeoutError as e:
-                if self.source.subscribe:
-                    return None
-                else:
-                    raise e
+            #        raise IndexError(f'Input sources indices misaligned or not equal! {row["_index_relative"]} =! {part_row["_index_relative"]}')
+                #print(part_row)
+                #TODO REMOVE HOT FIX BECAUSE CSV DATA IS USING HOURS NOT SECONDS!!!
+                #if '_index' in part_row and source.source_type == 'csv':
+                #    part_row['_index'] *= 3600
 
             row.update(part_row)
-
+            #else:
+            #    print('pr: ', part_row)
+            #    raise ValueError('!')
+        #row.pop('_index')
         row.update({'_index': t, '_index_relative': t - self.t0})
-
+        #print('im row: ', row)
         return row
 
+
 class RepeatedFunction:
     def __init__(self, interval, function, run_initially=False, *args, **kwargs):
         self._timer = None
@@ -1551,7 +1584,7 @@ class NumerousAdminClient(NumerousClient):
             yield json.loads(doc.json)
 
     def update_execution(self, update_):
-        self._base_client.UpdateExecution(spm_pb2.Json(json=json.dumps(update_)))
+        self._base_client.UpdateExecution(spm_pb2.Json(json=json.dumps(update_, default=datetime_to_json_converter)))
 
     def get_execution_status(self, exe_id):
         return self._job_manager.GetExecutionStatus(spm_pb2.ExecutionId(execution_id=exe_id)).json
@@ -1822,6 +1855,10 @@ class NumerousAdminClient(NumerousClient):
 
         ))
 
+    def get_cluster_info(self):
+
+        return self._job_manager.GetClusterInfo(spm_pb2.ClusterInfoRequest())
+
 @contextmanager
 def open_admin_client(server=None, port=None, secure=None, refresh_token=None) -> NumerousAdminClient:
     numerous_client_ = NumerousAdminClient(server=server, port=port, secure=secure, refresh_token=refresh_token)
